-- $Id: BaffleSet_03.lua,v 1.2 2007/07/30 23:41:54 dj Exp $

$ifnot BaffleSet

local errpfx = 'BaffleSet_03: '

BaffleSet = {}

SAOTRACE_LUALIB = getenv('SAOTRACE_LUALIB')

if ( nil == SAOTRACE_LUALIB ) then
   error ( errpfx .. 'environmental variable SAOTRACE_LUALIB is not set.' )
end

function dofilex( file )

   if ( nil == dofile( file ) ) then
      error( %errpfx .. 'error reading ' .. file )
   end

end

dofilex( SAOTRACE_LUALIB .. '/RDB.lua' )
dofilex( SAOTRACE_LUALIB .. '/suplib/tables.lua' )
dofilex( SAOTRACE_LUALIB .. '/suplib/paths.lua' )
dofilex( SAOTRACE_LUALIB .. '/suplib/strings.lua' )



if ( nil == SAOTRACE_DB ) then
  error (  errpfx .. 'environmental variable SAOTRACE_DB is not set.' )
end

dofilex( SAOTRACE_DB .. '/apertures/BaffleFuncs_02.lua' )


--------------------------------------------------------------------------
-- data types
--
--   displacement {
--     dx              -- (mm)           decenter
--     dy              -- (mm)           decenter
--     dz              -- (mm)           despace
--   }
--   disorientation {
--     azmis           -- (radians)
--     elmis           -- (radians)
--     clock           -- (radians)
--   }
--
--  annulus_info {
--      ri             -- (mm)       annulus inner radius
--      ro             -- (mm)       annulus outer radius
--      dri            -- (mm)       tolerance on inner radius
--      dro            -- (mm)       tolerance on outer radius
--      type           -- -1 = no annulus, 0 = transp, 1 = opaque
--  }
--
--  strut_info {
--      num            --            number of struts
--      width          -- (mm)       strut width
--      theta0         -- (radians)  offset of first strut
--      delta_theta    -- (radians)  angle between strut centerlines
--      theta0         -- (radians)  offset of first strut
--      R_i            -- (mm)       innermost radius for strut
--      R_o            -- (mm)       outermost radius for strut
--      type           -- strut | rect
--  }
--
--  baffle { 
--      id             --            identifier
--      z              -- (mm)       axial position of baffle
--      annulus_info   -- table      annulus data
--      strut_info     -- table      strut data
--      displacement   -- table      position    tweaks
--      disorientation -- table      orientation tweaks
--  }


--------------------------------------------------------------------------
-- BaffleSet:new
--
-- create a new baffle set.
--
-- if called with arguments
--    prog, rdbfile, shell, coord_type
-- then the baffle descriptions are read in from rdbfile

function BaffleSet:new( ... )

  local self = table_copy( BaffleSet )

  if ( arg.n > 0 ) then
    self:read_config( arg[1], arg[2], arg[3], arg[4] )
                              --    assembly_name, config_db, shell, args
  end

  return self
end

--------------------------------------------------------------------------
-- BaffleSet:read_config
--
-- read baffle assembly descriptions from an rdbfile;
-- construct a baffle assembly
--
-- Inputs:
--
--   assembly_name  - name of the assembly
--   shell          - shell number to be constructed
--   config_db      - rdb file containing information on how to
--                    construct the assemblies
--   arg            - table of optional parameters:
--        coord_type     - coordinate conversion type ('HRMA' or 'XRCF')
--                         used in converting the database values to
--                         raytrace coordinates. defaults to 'XRCF'
--        insert_struts  - insert struts?  defaults to true.
--        tol_scale      - tolerance scale.  see apply_annulus_tolerances()

function BaffleSet:read_config( assembly_name, config_db, shell, arg )

   if ( nil == assembly_name ) then
      error( "BaffleSet:read_config: assembly_name not set" )
   end

   local errpfx = "BaffleSet:read_config: " .. assembly_name .. ': '

   if ( nil == config_db ) then
      error( errpfx .. 'config_db not set' )
   end

   if ( nil == shell ) then
      error( errpfx .. 'shell not set' )
   end

   self.assembly_name = assembly_name
   self.shell         = shell

   self.coord_type = strupper( arg.coord_type or 'XRCF' )

   self.insert_struts = arg.insert_struts
   if ( nil == self.insert_struts ) then
      self.insert_struts = 1
   end

   self.coord_rot  = 0.0
   if ( self.coord_type == 'XRCF' ) then
      self.coord_rot = 180.0 * BaffleFuncs:deg2rad()
   end
    
   local rdb = RDB:new( config_db )
   if ( nil == rdb ) then
      error( errpfx .. 'unable to open data file: ' .. config_db )
   end

   local match = nil
   local data = rdb:read( )

   while ( not match and data ) do
      if ( self.assembly_name == data.id ) then
	 match = 1
      else
	 data = rdb:read( )
      end
   end

   if ( not match ) then
      error( errpfx .. "couldn't find assembly '" .. id .. "' in " .. config_db )
   end
   
   self.tol_scale      = data.tol_scale
   self.assembly_arg   = data.assembly_arg
   
   self.displacement   = 
      BaffleFuncs.construct_displacement( 
					 data.delta_X,
					 data.delta_Y,
					 data.delta_Z 
				      )
   
   self.disorientation = 
      BaffleFuncs.construct_disorientation( 
					   data.azmis_H,
					   data.elmis_H,
					   data.clock_H
					)
   
   -- interpolate any possible environmental variables
   self.baffle_cfg = str_interp( data.baffle_cfg )
   
   -- if baffle_cfg is not absolute, prepend directory of config_db
   if ( "/" ~= strsub( self.baffle_cfg, 1, 1) ) then
      local dirname, basename = splitpath( config_db )
      
      self.baffle_cfg = dirname .. "/" .. self.baffle_cfg
   end
   
   -- read in baffle information; construct baffle set
   --
   self:read_baffle_info()
   
   return self
   
end

--------------------------------------------------------------------------
-- BaffleSet:read_baffle_info
--
-- read baffle assembly descriptions from an rdbfile;
-- construct a baffle assembly
--

function BaffleSet:read_baffle_info()

   local errpfx = 'BaffleSet:read_baffle_info: ' .. self.assembly_name .. ': '

   local rdb = RDB:new( self.baffle_cfg )
   if ( nil == rdb ) then
      error( errpfx .. 'unable to open data file:'  .. self.baffle_cfg )
   end

   -- find the entry which corresponds to this shell


   -- Record: z - axial coordinate, ri - inner radius, ro - outer radius
   -- Dimensions are in mm; in raytrace coordinates.

   self.first_baffle = 1
   self.last_baffle  = 0

   self.baffle  = {}
   local baffle = {}

   local pid = 0
   local match = nil
   local data = rdb:read( )
   
   while ( data ) do

      if ( shell == data.shell ) then

	 pid = pid + 1

	 local baffle = {}

	 baffle.id = data.id
	 baffle.z  = -data.X

	 baffle.displacement =
	    BaffleFuncs.construct_displacement( data.delta_X, 
					        data.delta_Y, 
					        data.delta_Z )
	 baffle.disorientation =
	    BaffleFuncs.construct_disorientation( data.azmis_H, 
                                                  data.elmis_H, 
                                                  data.clock_H )

	 baffle.annulus_info = BaffleFuncs.construct_annulus( data )

	 if ( not self.insert_struts or nil == data.strut_type ) then
	    data.strut_type = 'none'
	 end

	 baffle.strut_info = BaffleFuncs.construct_struts( data )

	 self.baffle[pid] = baffle

      end

      data = rdb:read( )

   end

   self.last_baffle = pid

   if 0 == pid then
      error( errpfx .. "couldn't find shell=" .. shell .. " in " .. self.baffle_cfg )
   end

   self:apply_annulus_tolerances()

   -----------------------------------------
   -- body center z in raytrace coordinates:
   
   self.z_center = (self.baffle[self.first_baffle].z 
		    + self.baffle[self.last_baffle].z)/2.0

   pid = self.first_baffle
   while ( pid <= self.last_baffle ) do
      self.baffle[pid].z = self.baffle[pid].z - self.z_center
      pid = pid + 1
   end

   return self
end


--------------------------------------------------------------------------
-- BaffleSet.apply_annulus_tolerances
--
-- Inputs:
--
-- apply tolerances to baffle radii, scaled by tol_scale.
--    tol_scale = -1   -- minimum ghosting, maximum vignetting
--                 0   -- nominal baffles; zero tolerances applied
--                 1   -- maximum ghosting, minimum vignetting
-- Nonintegral values for tol_scale are permitted.

function BaffleSet:apply_annulus_tolerances()

  pid = self.first_baffle
  while ( pid <= self.last_baffle ) do

    if ( self.baffle[pid].annulus_info.type >= 0 ) then

       self.baffle[pid].annulus_info.dri = self.baffle[pid].annulus_info.dri * self.tol_scale
       self.baffle[pid].annulus_info.dro = self.baffle[pid].annulus_info.dro * self.tol_scale

       self.baffle[pid].annulus_info.ri 
                     = self.baffle[pid].annulus_info.ri 
                     - self.baffle[pid].annulus_info.dri

       self.baffle[pid].annulus_info.ro 
                     = self.baffle[pid].annulus_info.ro 
                     + self.baffle[pid].annulus_info.dro

    end
    pid = pid + 1

  end


end

--------------------------------------------------------------------------
-- BaffleSet.construct_baffles
--
-- Inputs:
--
--  debug_baffle -- id of baffle to debug, or 'all' for whole set

function BaffleSet:construct_baffles( debug_baffle )

  -- adjust position of assembly as a whole
  --
  -- push_CTM()               -- push current transformation matrix

  -- rotate according to the raytrace system in use.  the default
  -- is to map hrma to raytrace; xrcf to raytrace is rotated about
  -- the system mechanical axis by 180 degrees.

  rotate_z( self.coord_rot )

  -- the baffle set is specified in body centered coordinates
  -- move it to the correct position in raytrace coordinates
  -- note that z_center is the position of the body
  -- center in raytrace coordinates.

  translate(0,0, self.z_center)

  BaffleFuncs.apply_displacement_disorientation( self.displacement,
                                                 self.disorientation )

  -- print_assembly()           -- for tests only

  local id = self.first_baffle

  while id <= self.last_baffle do

    --> pushctm()

    BaffleFuncs.apply_displacement_disorientation( 
                         self.baffle[id].displacement,
                         self.baffle[id].disorientation )
      -- tweak the baffle

    begin_assembly( self.assembly_arg )

      -- print_photon( 'raw_project' )

      BaffleFuncs.do_baffle( self.baffle[id], debug_baffle )

      -- if ( self.baffle[id].id == debug_baffle or debug_baffle == "all" ) then
      --   print_assembly()
      -- end

    end_assembly()

    --> popctm()

    id = id + 1

  end

end

--------------------------------------------------------------------------
-- BaffleSet:get_body_baffleset_center_axial_station
--
-- return body center axial station of baffleset as a whole

function BaffleSet:get_baffleset_body_center_axial_station()

  return self.z_center

end

--------------------------------------------------------------------------
-- BaffleSet:get_baffleset_displacement
--
-- return displacement of baffleset as a whole

function BaffleSet:get_baffleset_displacement()

  return self.displacement

end

--------------------------------------------------------------------------
-- BaffleSet:get_baffleset_disorientation
--
-- return disorientation of baffleset as a whole

function BaffleSet:get_baffleset_disorientation()

  return self.disorientation

end

--------------------------------------------------------------------------
-- BaffleSet:get_first_baffle_number
--
-- return id number of first baffle

function BaffleSet:get_first_baffle_number()

  return self.first_baffle

end

--------------------------------------------------------------------------
-- BaffleSet:get_last_baffle_number
--
-- return id number of last baffle

function BaffleSet:get_last_baffle_number()

  return self.last_baffle

end

--------------------------------------------------------------------------
-- BaffleSet:get_baffle_id
--
-- return id for baffle pid
--
-- Inputs:
--
--   pid -- baffle index number

function BaffleSet:get_baffle_id( pid )

  return self.baffle[pid].id

end

--------------------------------------------------------------------------
-- BaffleSet:get_baffle_net_displacement
--
-- return net displacement of baffle pid
-- (includes displacement of baffle set as a whole)
--
-- Inputs:
--
--   pid -- baffle index number

function BaffleSet:get_baffle_net_displacement( pid )

  local dsp = self.baffle[pid].displacement
  dsp.dx = dsp.dx + self.displacement.dx
  dsp.dy = dsp.dy + self.displacement.dy
  dsp.dz = dsp.dy + self.displacement.dz
  return dsp

end

--------------------------------------------------------------------------
-- BaffleSet:get_baffle_displacement
--
-- return displacement of baffle pid
--
-- Inputs:
--
--   pid -- baffle index number

function BaffleSet:get_baffle_displacement( pid )

  return self.baffle[pid].displacement

end

--------------------------------------------------------------------------
-- BaffleSet:get_baffle_disorientation
--
-- return disorientation of baffle pid
--
-- Inputs:
--
--   pid -- baffle index number

function BaffleSet:get_baffle_disorientation( pid )

  return self.baffle[pid].disorientation

end

--------------------------------------------------------------------------
-- BaffleSet:get_baffle_axial_station
--
-- return axial station of baffle pid
-- NB:  does not include the effect of any baffle displacements or
--      disorientation
--
-- Inputs:
--
--   pid -- baffle index number

function BaffleSet:get_baffle_axial_station( pid )

  return self.baffle[pid].z + self.z_center

end


--------------------------------------------------------------------------
-- BaffleSet:get_annulus_info
--
-- return annulus_info for baffle pid
--
-- Inputs:
--
--   pid -- baffle index number

function BaffleSet:get_annulus_info( pid )

  return self.baffle[pid].annulus_info

end



--------------------------------------------------------------------------
-- BaffleSet:get_strut_info
--
-- return strut_info for baffle pid
--
-- Inputs:
--
--   pid -- baffle index number

function BaffleSet:get_strut_info( pid )

  return self.baffle[pid].strut_info

end

$end
